////////////////////////////////////////
///@file rx_pkt_sample.c
///
///@brief pag_get_pkt接口收包示例代码，收包线程可以指定收包个数和打印报文内容
///
/// 修改记录:
///
///@author lifengw@sugon.com
///@date 2011-12-21
///
/// 修改内容：建立文档
////////////////////////////////////////

//系统路径下头文件
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>  //getopt
#include <string.h>
#include <pthread.h>
#include <signal.h>
#include <sys/time.h>
#include <linux/types.h>
#include <netinet/ip.h> //struct ip
#include <netinet/tcp.h>
#include <netinet/udp.h>
#include <netinet/ip6.h>
#include <net/ethernet.h>  //struct ether_header

#include <libpag.h>

//本地路径下包含文件

///// 宏定义 ///////

#define DEF_THREADS_NUM 4   ///<默认线程数
#define MAX_THREADS_NUM 64  ///<系统支持的最大线程数 
#define PKT_HEADER_LEN 50 ///<报文附加包头的长度sizeof(struct pkt_header)

#ifndef ETHERTYPE_IPV6
#define ETHERTYPE_IPV6 0x86dd
#endif

#ifndef ETHERTYPE_VLAN
#define ETHERTYPE_VLAN 0x8100
#endif

///// 数据结构定义 ///////

////////// 数据 /////////

static int glob_loop_flag_main;
static int glob_loop_flag_thread;
static int glob_copy_time;
static int glob_print_pkt_flag;
static int glob_pkts_num;
static int glob_threads_num;
static int glob_compare_content_flag;
static int glob_compare_sid_flag;
static int glob_pktsend_flag;
static int glob_mode;
static int glob_mode_flag;

///计数器，每个线程的计数器内存空间分开，以避免cache抖动
ulong glob_pkts[MAX_THREADS_NUM][16];
ulong glob_bytes[MAX_THREADS_NUM][16];
static int send_type[MAX_THREADS_NUM][16];

///////   函数列表  ///////////////
///mode 等于0表示，pdata指向NULL；等于15表示，pdata指向二层头；
///@param mod 等于31表示，pdata指向VLAN头；等于47表示，pdata指向三层头；
///@param mod 等于63表示，pdata指向四层头；等于79表示，pdata指向负载位置。
///@brief 输出程序帮助信息
///@param cmd 程序命令(输入)
void usage(char* cmd) 
{
    printf("Usage: %s [-m threads_num] [-n pkt_num] [-c copy_time] [-s] [-p] [-t test_bitmap] \
    [-o mode_value] [-O print pdata] [-?]\n", cmd);
    printf("  -m threads number, default = %d\n", DEF_THREADS_NUM);
    printf("  -n pkts num to rx, default = continously\n");
    printf("  -c copy time of each rx pkt, default = 0\n");
    printf("  -s send rcv pkt\n");
    printf("  -p print pkts and headers\n");
    printf("  -t compare bitmap, 1:content for pkt (byte increase) test, 2 for stream id test\n");
    printf("  -o mode=0,pdata=NULL;mode=15,pdata=mac;mode=31,pdata=vlan;mode=47,pdata=ip;\n");
    printf("     mode=63,pdata=tdp;mode=79,pdata=payload\n");
    printf("  -O print pdata\n");
    printf("  -? print usage message\n");
    printf("\n");
}

///@brief 打印收包信息的线程，每隔1秒显示一次
///@param ptr 线程建立时的指针，这里没有使用
void print_speed_thread(void *ptr)
{
    int i;
    ulong old_pkts[MAX_THREADS_NUM];
    ulong old_bytes[MAX_THREADS_NUM];

    for (i = 0; i < glob_threads_num; i++)
    {
        old_pkts[i] = 0;
        old_bytes[i] = 0;
    }
    
    while (glob_loop_flag_thread)
    {
        sleep(1); 
        printf("\nCurrent speed----- ");
        printf("\n");        
        for (i = 0; i < glob_threads_num; i++) 
        {
            printf("\t Thread[%d]:\t %8lu pps, %8lu M bps\n", i, 
            glob_pkts[i][0] - old_pkts[i], 
            (glob_bytes[i][0] - old_bytes[i]) >> 17);
            old_pkts[i] = glob_pkts[i][0];
            old_bytes[i] = glob_bytes[i][0];
        }
    }
}


///@brief 拷贝报文（模拟应用处理包的负载）
///@param pkt 指向报文的指针
///@param len 报文长度
inline void copy_pkt(void *pkt, int len)
{
    int i, total_len;
    uint8_t *p8;
    uint8_t p8_tmp[2000];
    
    p8 = (uint8_t *)pkt;
    
    ///拷贝报文
    for (i = 0; i < len; i++) 
    {
        p8_tmp[i] = p8[i];
    }
}

///@brief 打印报文
///@param pkt 指向报文的指针
///@param len 报文长度
static inline void print_pkt(void *pkt, int len)
{
    int i;
    uint8_t *p8;
    static int pkt_print;
    int total_length;
 
    pkt_print++;
    printf("\n====  pkt:%d  ==== len = %d\n", pkt_print , len);
    
    ///找到报文附加头, 打印长度包括附加头
    ///假定收到的报文都是标准的以太网报文
    total_length = len + sizeof(struct ether_header) + PKT_HEADER_LEN;
    p8 = (uint8_t *)pkt - sizeof(struct ether_header) - PKT_HEADER_LEN;
    
    ///打印报文, 每16字节一行，128字节一段
    for (i = 0; i < total_length; i++)
    {
        if ((i % 128 == 0))
        {
            printf("\n");
        }
        if ((i % 16 == 0))
        {
            printf("\n%03x: ", i / 16 + 1);
        }
        if (i % 8 == 0)
        {
            printf(" ");
        }
        printf("%02x ", p8[i]);
    }
    printf("\n");
}

///@brief 获取二层头长度
int get_frame_length(void *pdata)
{
    uint8_t *p8;
    uint16_t *p16;
    int len;
    
    p8 = pdata;
    
    ///判断是否是标准的eth头，14个字节
    p16 = (uint16_t *)(p8 + (ETHER_ADDR_LEN << 1));
    if ((*p16 == ETHERTYPE_IP) || (*p16 == ETHERTYPE_IPV6))
    {
        len = sizeof(struct ether_header);
    }
    
    ///判断是否是vlan报文，包括一重vlan和双重vlan
    if ((*p16 == ETHERTYPE_VLAN) && ((*(p16 + 2) == ETHERTYPE_IP) || (*(p16 + 2) == ETHERTYPE_IPV6)))
    {
        len = sizeof(struct ether_header) + 4;
    }
    
    if ((*p16 == ETHERTYPE_VLAN) && (*(p16 + 2) == ETHERTYPE_VLAN) && \
        ((*(p16 + 2) == ETHERTYPE_IP) || (*(p16 + 2) == ETHERTYPE_IPV6)))
    {
        len = sizeof(struct ether_header) + 8;
    }
    
    return len;
}

///@brief 获取VLAN头长度
int get_vlan_length(void *pdata)
{
    uint16_t *p16;
    int len;
    
    p16 = (uint16_t *)pdata;
    
    ///判断是否是vlan报文，包括一重vlan和双重vlan
    if ((*p16 == ETHERTYPE_VLAN) && ((*(p16 + 2) == ETHERTYPE_IP) || (*(p16 + 2) == ETHERTYPE_IPV6)))
    {
        len = 4;
    }
    
    if ((*p16 == ETHERTYPE_VLAN) && (*(p16 + 2) == ETHERTYPE_VLAN) && \
        ((*(p16 + 2) == ETHERTYPE_IP) || (*(p16 + 2) == ETHERTYPE_IPV6)))
    {
        len = 8;
    }
    
    return len;
}

///获取v4的传输层数据长度
int get_tdp_length_v4(void *pkt, int len)
{
    struct ip *iph;
    int length;
    
    iph = (struct ip *)pkt;
    length = len - (iph->ip_hl << 2);
    
    return length;
}

///获取v6的传输层数据长度
int get_tdp_length_v6(void *pkt, int len)
{
    struct ip6_hdr *iph6;
    int current_length, tdp_len;
    int nh;
    
    iph6 = (struct ip6_hdr *)pkt;
    current_length = sizeof(struct ip6_hdr);
    nh = iph6->ip6_nxt;
    
    switch (nh)
    {
        case IPPROTO_TCP:
            tdp_len = len - current_length;
            break;
        case IPPROTO_UDP:
            tdp_len = len - current_length;
            break;
    }

    return tdp_len;
}

///获取传输层数据长度
int get_tdp_length(void *pkt, int len, int version)
{
    struct ip *iph;
    struct ip6_hdr *iph6;
    int tdp_len;
    
    if (version == 0x04)
    {
        iph = (struct ip *)pkt;
        if (iph->ip_p == IPPROTO_TCP || iph->ip_p == IPPROTO_UDP)
        {
            tdp_len = get_tdp_length_v4(pkt, len);
        }
    }
    else if (version == 0x06)
    {
        iph6 = (struct ip6_hdr *)pkt;
        if (iph6->ip6_nxt == IPPROTO_TCP || iph6->ip6_nxt == IPPROTO_UDP)
        {
            tdp_len = get_tdp_length_v6(pkt, len);
        }
    }
    
    return tdp_len;
}

///获取v4的payload数据长度
int get_pld_length_v4(void *pkt, int len)
{
    struct ip *iph;
    struct tcphdr *tcph;
    struct udphdr *udph;
    int pld_len;
    
    iph = (struct ip *)pkt;
    if (iph->ip_p == IPPROTO_TCP)
    {
        tcph = (struct tcphdr *)(pkt + (iph->ip_hl << 2));
        pld_len = len - (iph->ip_hl << 2) - (tcph->doff << 2);
    }
    else if (iph->ip_p == IPPROTO_UDP)
    {
        udph = (struct udphdr *)(pkt + (iph->ip_hl << 2));
        pld_len = udph->len - sizeof(struct udphdr);
    }
    
    return pld_len;
}

///获取v6的payload数据长度
int get_pld_length_v6(void *pkt, int len)
{
    struct ip6_hdr *iph6;
    struct tcphdr *tcph;
    struct udphdr *udph;
    int pld_len;
    
    iph6 = (struct ip6_hdr *)pkt;
    if (iph6->ip6_nxt == IPPROTO_TCP)
    {
        tcph = (struct tcphdr *)(pkt + sizeof(struct ip6_hdr));
        pld_len = ntohs(iph6->ip6_plen) - (tcph->doff << 2);
    }
    else if (iph6->ip6_nxt == IPPROTO_UDP)
    {
        udph = (struct udphdr *)(pkt + sizeof(struct ip6_hdr));
        pld_len = ntohs(iph6->ip6_plen) - sizeof(struct udphdr);
    }
    
    return pld_len;
}

///获取payload数据长度
int get_pld_length(void *pkt, int len, int version)
{
    struct ip *iph;
    struct ip6_hdr *iph6;
    int tdp_len;
    
    if (version == 0x04)
    {
        iph = (struct ip *)pkt;
        if (iph->ip_p == IPPROTO_TCP || iph->ip_p == IPPROTO_UDP)
        {
            tdp_len = get_pld_length_v4(pkt, len);
        }
    }
    else if (version == 0x06)
    {
        iph6 = (struct ip6_hdr *)pkt;
        if (iph6->ip6_nxt == IPPROTO_TCP || iph6->ip6_nxt == IPPROTO_UDP)
        {
            tdp_len = get_pld_length_v6(pkt, len);
        }
    }
    
    return tdp_len;
}

///根据mode值打印相应报文段
void print_pdata(void *pkt, int len, void *pdata, uint8_t version)
{
    uint8_t *p8;
    int i, total_len;
    
    printf("from print pdata...\n");
    switch (glob_mode)
    {
        case 0:
            return;
        case 15:
            total_len = len + get_frame_length(pdata);
            break;
        case 31:
            total_len = len + get_vlan_length(pdata);
            break;
        case 47:
            total_len = len;
            break;
        case 63:
            total_len = get_tdp_length(pkt, len, version);
            break;
        case 79:
            total_len = get_pld_length(pkt, len, version);
            break;
    }
    
    p8 = (uint8_t *)pdata;
    ///打印报文, 每16字节一行，128字节一段
    for (i = 0; i < total_len; i++)
    {
        if ((i % 128 == 0))
        {
            printf("\n");
        }
        if ((i % 16 == 0))
        {
            printf("\n%03x: ", i / 16 + 1);
        }
        if (i % 8 == 0)
        {
            printf(" ");
        }
        printf("%02x ", p8[i]);
    }
    printf("\n");
}

///@brief 检查收包内容
///@param pkt 指向报文的指针
///@param len 报文的长度
void compare_payload_content(void *pkt)
{
    int i;
    int payload_len;
    __u8 *p8;
    static __u64 pkts_num;
    
    pkts_num++;
    p8 = (__u8 *)pag_get_payload(pkt);
    payload_len = pag_get_payload_length(pkt);
    
    ///比较负载内容
    for (i = 0; i < payload_len; i++)
    {
        if(p8[i] != (__u8)i)
        {
            printf("pkts = %d, i = %d, payload = %d\n", pkts_num, i, p8[i]);
        }
    }
}

///@brief 检查收包流号是否正确
///@param pkt 指向报文ip头的指针
///@param sid 接收到报文的流号
void compare_sid(void *pkt, int sid)
{
    static __u64 pkts_num;
    struct tuple4 t4;
    struct iphdr *pip;
    struct tcphdr *ptcp;
    int s;
    uint8_t nl_proto;
    
    
    pkts_num++;
    pip = (struct iphdr *)pkt;
    ptcp = (struct tcphdr *)((void *)pip + pip->ihl * 4);
    
    t4.saddr = pip->saddr;
    t4.daddr = pip->daddr;
    t4.source = ptcp->source;
    t4.dest = ptcp->dest;
    
    s = get_pkt_stream_no(&t4);
    if (sid != s)
    {
        printf("pkts = %d, should in stream [%d] recieved by %d\n", pkts_num, s, sid);
    }
}

///@brief 转发接收到的数据报文
///@param pkt 指向报文ip头的指针
///@param sid 接收到报文的流号
void send_rcvpkt(void *pkt, int sid, int len)
{
    struct iphdr *iph;
    void *buff;
    
    ///获得发包缓冲区
    buff = pag_getsendbuf(sid);
    while (!buff) 
    {
        usleep(500);
        buff = pag_getsendbuf(sid);
    }
    
    ///拷贝报文
    memcpy(buff, pkt, len);
    
    ///发送报文
    pag_send(send_type[sid][0], sid, len);
    
    ///切换发包类型
    send_type[sid][0] = !send_type[sid][0];
}

///@brief 接收报文的线程
///@param ptr 线程建立时的指针，这里指向线程的数据流号
void rx_thread(void *ptr)
{
    int i, sid;
    void *pkt, *pdata, *ph;
    struct ip *iph;
    struct ip6_hdr *iph6;
    int len;
    uint8_t version;

    sid = *((int *)ptr);
    
    ///循环，直到线程循环标志清零   
    while (glob_loop_flag_thread)
    {
        pkt = pag_get_pkt(sid, glob_mode, &pdata);
        
        ///--如果没有报文，则循环等待
        if (!pkt) 
        {
            usleep(500);
            continue;
        }
        
        ///--增加报文计数
        glob_pkts[sid][0]++;
        
        ///--获取ip包
        version = *(uint8_t *)pkt;
        version >>= 4;
        if (version == 0x04)
        {
            iph = (struct ip *)pkt;
            len = ntohs(iph->ip_len);
            ph = iph;   
        }
        else if (version == 0x06)
        {
            iph6 = (struct ip6_hdr *)pkt;
            len = ntohs(iph6->ip6_plen) + sizeof(struct ip6_hdr);
            ph = iph6;  
        }
        
        glob_bytes[sid][0] += len;
        
        ///--根据参数对报文做相应处理
        for (i = 0; i < glob_copy_time; i++)
        {
            copy_pkt(pkt, len);
        }
        
        if (glob_print_pkt_flag)
        {
            print_pkt(pkt, len);
        }
        
        if (glob_mode_flag)
        {
            print_pdata(pkt, len, pdata, version);    
        }
        
        if (glob_compare_content_flag == 1)
        {
            //compare_payload_content(pkt);
        }
        
        if (glob_compare_sid_flag == 1)
        {
            //compare_sid(pkt, sid);
        }
        
        if (glob_pktsend_flag == 1)
        {
            send_rcvpkt((void *)ph, sid, len);
        }
    }
}

///@brief 信号响应函数，收到信号后把全局循环标志无效掉
void stop_loop(int sig)
{
    glob_loop_flag_main = 0;
}

///@brief 主函数
///@return 成功返回0，失败返回-1
int main(int argc, char *argv[])
{
    int c;
    struct timeval ts, te;
    double sec; 
    ulong all_pkts, all_bytes;
    sigset_t set;
    pthread_t rx_tids[MAX_THREADS_NUM];
    int sids[MAX_THREADS_NUM];
    pthread_t print_tid;
    int i;
    int bitmap;
    
    ///打印使用信息
    usage(argv[0]);

    ///读取参数
    glob_copy_time = 0;
    glob_print_pkt_flag = 0;
    glob_pkts_num = 0;
    glob_threads_num = DEF_THREADS_NUM;
    glob_pktsend_flag = 0;
    glob_mode = 0;
    glob_mode_flag = 0;
    
    while ((c = getopt(argc, argv, "c:n:m:t:o:Osp?")) != -1) 
    {
        switch (c)
        {
            case 'c':
                glob_copy_time = atoi(optarg);
                break;
            case 'n':
                glob_pkts_num = atoi(optarg);
                break;
            case 'm':
                glob_threads_num = atoi(optarg);
                if (glob_threads_num > MAX_THREADS_NUM)
                {
                    printf("\nError: threads num invalid\n");
                    return -1;
                }
                break;
            case 't':
                bitmap = atoi(optarg);
                if (bitmap & 1)
                {
                    glob_compare_content_flag = 1;
                }
                else if (bitmap & 2)
                {
                    glob_compare_sid_flag = 1;
                }
                else
                {
                    printf("\nError: test bitmap invalid\n");
                    return -1;
                }
                break;
            case 'o':
                glob_mode = atoi(optarg);
                break;
            case 'O':
                glob_mode_flag = 1;
                break;
            case 'p':
                glob_print_pkt_flag = 1;
                break;
            case 's':
                glob_pktsend_flag = 1;
                break;
            case '?':
            default:
                usage(argv[0]);
                return 0;
        }
    }
    
    ///打开设备
    if (pag_open() < 0)
    {
        printf("### Error : can not open device! ###\n");
        return -1;
    }

    ///创建线程
    glob_loop_flag_thread = 1;
    for (i = 0; i < glob_threads_num; i++)
    {
        sids[i] = i;
        if (pthread_create(&rx_tids[i], NULL, (void *)rx_thread, 
            (void *)&sids[i]) != 0)   
        {
            printf("### Error: Can not creat threads!\n");    
            pag_close();
            return -1;
        }
    }
    
    ///如果不打印报文内容，则打印报文速度
    if (!glob_print_pkt_flag)
    {
        if (pthread_create(&print_tid, NULL, (void *)print_speed_thread,
            NULL) != 0)   
        {
            printf("### Error: Can not creat threads!\n");    
            pag_close();
            return -1;
        }
    }

    ///等待结束信号，或着收到期望的包数 
    sigemptyset(&set);
    sigprocmask(SIG_SETMASK, &set, NULL);
    signal(SIGINT, stop_loop);
    
    gettimeofday(&ts, NULL);   
     
    glob_loop_flag_main = 1;
    while (glob_loop_flag_main) 
    {
        usleep(10);  
        
        all_pkts = 0;
        all_bytes = 0;
        for (i = 0; i < glob_threads_num; i++) 
        {
            all_pkts += glob_pkts[i][0];
        }
        
        if (glob_pkts_num != 0 && glob_pkts_num <= all_pkts)
        {
            break;
        }
    }
            
    gettimeofday(&te, NULL);  
    
    ///停止线程
    glob_loop_flag_thread = 0;
    
    for (i = 0; i < glob_threads_num; i++)
    {
        pthread_join(rx_tids[i], NULL);
    }    
    if (!glob_print_pkt_flag)
    {
        pthread_join(print_tid, NULL);
    }
    
    ///关闭设备
    pag_close();    
    
    ///打印总结信息
    sec = (float)(te.tv_sec - ts.tv_sec) + ((float)(te.tv_usec - ts.tv_usec)) 
        / 1000000.f;   
    printf("\n");        
    printf("Pkts recieved in threads: \n");
    all_pkts = 0;
    all_bytes = 0;
    for(i=0; i<glob_threads_num; i++) 
    {
        printf("\t Thread[%d]:\t <%8lu> pkts, <%8lu M> bytes\n", i,
            glob_pkts[i][0], glob_bytes[i][0] >> 20);
        all_pkts += glob_pkts[i][0];
        all_bytes += glob_bytes[i][0];
    }
    printf("Total : %lu pkts,  %lu bytes\n", all_pkts, all_bytes);
    printf("Speed : < %u > pps, < %u M > bps\n", 
        (int)(all_pkts / sec), (int)((all_bytes >> 17) / sec));
    printf("\n");        

    return 0;
}
