////////////////////////////////////////
///@file rx_sample.c
///
///@brief pag接口收包示例代码，收包线程可以指定收包个数和打印报文内容
///
/// 修改记录:
///
///@author douxg@dawning.com.cn
///@date 2011-01-07
///
///将报文附加头大小改回56字节。
///
///@author douxg@dawning.com.cn
///@date 2011-01-04
///
///将报文附加头大小改为64字节。
///
///@author liuzhh@dawning.com.cn
///@date 2010-10-09
///
///增加流号判断参数
///
///@author liuzhh@dawning.com.cn
///@date 2009-11-10
///
/// 修改内容：建立文档
////////////////////////////////////////

//系统路径下头文件
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>  //getopt
#include <string.h>
#include <pthread.h>
#include <signal.h>
#include <sys/time.h>
#include <linux/types.h>
#include <netinet/ip.h> //struct ip
#include <netinet/ip6.h>
#include <linux/tcp.h>
#include <net/ethernet.h>  //struct ether_header

#include <libpag.h>

//本地路径下包含文件

///// 宏定义 ///////

#define DEF_THREADS_NUM 4   ///<默认线程数
#define MAX_THREADS_NUM 64  ///<系统支持的最大线程数 
#define PKT_HEADER_LEN 50 ///<报文附加包头的长度sizeof(struct pkt_header)

///// 数据结构定义 ///////

////////// 数据 /////////

static int glob_loop_flag_main;
static int glob_loop_flag_thread;
static int glob_copy_time;
static int glob_print_pkt_flag;
static int glob_pkts_num;
static int glob_threads_num;
static int glob_compare_content_flag;
static int glob_compare_sid_flag;
static int glob_pktsend_flag;

///计数器，每个线程的计数器内存空间分开，以避免cache抖动
ulong glob_pkts[MAX_THREADS_NUM][16];
ulong glob_bytes[MAX_THREADS_NUM][16];
static int send_type[MAX_THREADS_NUM][16];

///////   函数列表  ///////////////

///@brief 输出程序帮助信息
///@param cmd 程序命令(输入)
void usage(char* cmd) 
{
    printf("Usage: %s [-m threads_num] [-n pkt_num] [-c copy_time] [-s] [-p] [-t test_bitmap] [-?]\n", cmd);
    printf("  -m threads number, default = %d\n", DEF_THREADS_NUM);
    printf("  -n pkts num to rx, default = continously\n");
    printf("  -c copy time of each rx pkt, default = 0\n");
    printf("  -s send rcv pkt\n");
    printf("  -p print pkts and headers\n");
    printf("  -t compare bitmap, 1:content for pkt (byte increase) test, 2 for stream id test\n");
    printf("  -? print usage message\n");
    printf("\n");
}

///@brief 打印收包信息的线程，每隔1秒显示一次
///@param ptr 线程建立时的指针，这里没有使用
void print_speed_thread(void *ptr)
{
    int i;
    ulong old_pkts[MAX_THREADS_NUM];
    ulong old_bytes[MAX_THREADS_NUM];

    for (i = 0; i < glob_threads_num; i++)
    {
        old_pkts[i] = 0;
        old_bytes[i] = 0;
    }
    
    while (glob_loop_flag_thread)
    {
        sleep(1); 
        printf("\nCurrent speed----- ");
        printf("\n");        
        for (i = 0; i < glob_threads_num; i++) 
        {
            printf("\t Thread[%d]:\t %8lu pps, %8lu M bps\n", i, 
            glob_pkts[i][0] - old_pkts[i], 
            (glob_bytes[i][0] - old_bytes[i]) >> 17);
            old_pkts[i] = glob_pkts[i][0];
            old_bytes[i] = glob_bytes[i][0];
        }
    }
}

///@brief 拷贝报文（模拟应用处理包的负载）
///@param pkt 指向报文的指针
///@param len 报文长度
inline void copy_pkt(packet_t *pkt)
{
    int i;
    int total_length, frame_length;
    __u8 p8_tmp[2000], *p8;
    __u16 frame_type;
    
    ///找到报文附加头, 拷贝长度包括附加头
    p8 = (__u8 *)pkt;
    frame_type = pag_get_frame_type(pkt);
    frame_length = pag_get_frame_length(pkt);
    
    ///根据frame类型为len赋值
    if (frame_type == 0)
    {
        total_length = frame_length + PKT_HEADER_LEN;
    }
    else if (frame_type == 1)
    {
        total_length = frame_length + 10 + PKT_HEADER_LEN;
    }
    else if (frame_type == 2)
    {
        total_length = frame_length + 2 + PKT_HEADER_LEN;   
    }
    else
    {
        total_length = 200;   
    }
    
    ///拷贝报文
    for (i = 0; i < total_length; i++) 
    {
        p8_tmp[i] = p8[i];
    }
}

///@brief 打印报文
///@param pkt 指向报文的指针
///@param len 报文长度
static inline void print_pkt(void *pkt, int len)
{
    int i;
    __u8 *p8;
    static int pkt_print;
    int total_length, frame_length;
    __u16 frame_type;
 
    pkt_print++;
    printf("\n====  pkt:%d  ==== len = %d\n", pkt_print , len);
    
    ///找到报文附加头, 打印长度包括附加头
    p8 = (__u8 *)pkt;
    frame_type = pag_get_frame_type(pkt);
    frame_length = pag_get_frame_length(pkt);
    
    if (frame_type == 0)
    {
        total_length = frame_length + PKT_HEADER_LEN;
    }
    else if (frame_type == 1)
    {
        total_length = frame_length + 10 + PKT_HEADER_LEN;
    }
    else if (frame_type == 2)
    {
        total_length = frame_length + 2 + PKT_HEADER_LEN;
    }
    else
    {
        total_length = 200;   
    }

    ///打印报文, 每16字节一行，128字节一段
    for (i = 0; i < total_length; i++)
    {
        if ((i % 128 == 0))
        {
            printf("\n");
        }
        if ((i % 16 == 0))
        {
            printf("\n%03x: ", i / 16 + 1);
        }
        if (i % 8 == 0)
        {
            printf(" ");
        }
        printf("%02x ", p8[i]);
    }
    printf("\n");
}

///@brief 检查收包内容
///@param pkt 指向报文的指针
///@param len 报文的长度
void compare_payload_content(packet_t *pkt)
{
    int i;
    int payload_len;
    __u8 *p8;
    static __u64 pkts_num;
    
    pkts_num++;
    p8 = (__u8 *)pag_get_payload(pkt);
    payload_len = pag_get_payload_length(pkt);
    
    ///比较负载内容
    for (i = 0; i < payload_len; i++)
    {
        if(p8[i] != (__u8)i)
        {
            printf("pkts = %d, i = %d, payload = %d\n", pkts_num, i, p8[i]);
        }
    }
}

///@brief 检查收包流号是否正确
///@param pkt 指向报文ip头的指针
///@param sid 接收到报文的流号
void compare_sid(void *pkt, int sid)
{
    static __u64 pkts_num;
    struct tuple4 t4;
    struct iphdr *pip;
    struct tcphdr *ptcp;
    int s;
    uint8_t nl_proto;
    
    pkts_num++;
    pip = (struct iphdr *)pkt;
    ptcp = (struct tcphdr *)((void *)pip + pip->ihl * 4);
    
    t4.saddr = pip->saddr;
    t4.daddr = pip->daddr;
    t4.source = ptcp->source;
    t4.dest = ptcp->dest;
    
    s = get_pkt_stream_no(&t4);
    if (sid != s)
    {
        printf("pkts = %d, should in stream [%d] recieved by %d\n", pkts_num, s, sid);
    }
}

///@brief 转发接收到的数据报文
///@param pkt 指向报文ip头的指针
///@param sid 接收到报文的流号
void send_rcvpkt(void *pkt, int sid, int len)
{
    struct iphdr *iph;
    void *buff;
    
    ///获得发包缓冲区
    buff = pag_getsendbuf(sid);
    while (!buff) 
    {
        usleep(500);
        buff = pag_getsendbuf(sid);
    }
    
    ///拷贝报文
    memcpy(buff, pkt, len);
    
    ///发送报文
    pag_send(send_type[sid][0], sid, len);
    
    ///切换发包类型
    send_type[sid][0] = !send_type[sid][0];
}

///@brief 转发接收到的数据报文
///@param pkt 指向报文帧头的指针
void send_rcvframe(void *pkt, int sid)
{
    void *buff, *frame;
    int frame_len;
    
    frame = pag_get_frame(pkt);
    frame_len = pag_get_frame_length(pkt);
    
    ///获得发包缓冲区
    buff = pag_getsendbuf(sid);
    while (!buff) 
    {
        usleep(500);
        buff = pag_getsendbuf(sid);
    }
    
    ///拷贝报文
    memcpy(buff, frame, frame_len);
    
    ///发送报文
    pag_send_eth(sid, buff, frame_len);
}

///@brief 接收报文的线程
///@param ptr 线程建立时的指针，这里指向线程的数据流号
void rx_thread(void *ptr)
{
    int sid, i;
    void *ip;
    int len;
    packet_t *pkt;    
    uint8_t nl_proto, tl_proto;

    sid = *((int *)ptr);
    
    ///循环，直到线程循环标志清零   
    while (glob_loop_flag_thread)
    {
#ifndef _CONFIG_C27 
        pkt = pag_get_packet(sid);
#else
        pkt = pag_get_pkt(sid);
#endif

        ///--如果没有报文，则循环等待
        if (!pkt) 
        {
            usleep(500);
            continue;
        }
        
        ///--增加报文计数
        glob_pkts[sid][0]++;
        
        ///--获取网络层协议
        nl_proto = pag_get_nl_proto(pkt);
        if (nl_proto == 0xff)
        {
            printf("unknown network layer proto\n");
            continue;
        }
        
        tl_proto = pag_get_tl_proto(pkt);
        if (tl_proto != IPPROTO_TCP && tl_proto != IPPROTO_UDP)
        {
            printf("unknown transport layer proto\n");
            continue;   
        }
        
        ///--获取网络层数据
        ip = pag_get_ip(pkt);
        
        ///--获取网络层数据长度
        len = pag_get_ip_length(pkt);
        
        glob_bytes[sid][0] += len;
        
        ///--根据参数对报文做相应处理
        for (i = 0; i < glob_copy_time; i++)
        {
            copy_pkt(pkt);
        }
        
        if (glob_print_pkt_flag)
        {
            print_pkt(pkt, len);
        }
        
        if (glob_compare_content_flag == 1)
        {
            compare_payload_content(pkt);
        }
        
        if (glob_compare_sid_flag == 1)
        {
            //compare_sid(pkt, sid);
        }
                
        if (glob_pktsend_flag == 1)
        {
        #ifndef _CONFIG_C27
            send_rcvpkt((void *)ip, sid, len);
        #else
            send_rcvframe(pkt, sid);
        #endif
        }
    }
}

///@brief 信号响应函数，收到信号后把全局循环标志无效掉
void stop_loop(int sig)
{
    glob_loop_flag_main = 0;
}

///@brief 主函数
///@return 成功返回0，失败返回-1
int main(int argc, char *argv[])
{
    int c;
    struct timeval ts, te;
    double sec; 
    ulong all_pkts, all_bytes;
    sigset_t set;
    pthread_t rx_tids[MAX_THREADS_NUM];
    int sids[MAX_THREADS_NUM];
    pthread_t print_tid;
    int i;
    int bitmap;
    
    ///打印使用信息
    usage(argv[0]);

    ///读取参数
    glob_copy_time = 0;
    glob_print_pkt_flag = 0;
    glob_pkts_num = 0;
    glob_threads_num = DEF_THREADS_NUM;
    glob_pktsend_flag = 0;
    
    while ((c = getopt(argc, argv, "c:n:m:t:sp?")) != -1) 
    {
        switch (c)
        {
            case 'c':
                glob_copy_time = atoi(optarg);
                break;
            case 'n':
                glob_pkts_num = atoi(optarg);
                break;
            case 'm':
                glob_threads_num = atoi(optarg);
                if (glob_threads_num > MAX_THREADS_NUM)
                {
                    printf("\nError: threads num invalid\n");
                    return -1;
                }
                break;
            case 't':
                bitmap = atoi(optarg);
                if (bitmap & 1)
                {
                    glob_compare_content_flag = 1;
                }
                else if (bitmap & 2)
                {
                    glob_compare_sid_flag = 1;
                }
                else
                {
                    printf("\nError: test bitmap invalid\n");
                    return -1;
                }
                break;
            case 'p':
                glob_print_pkt_flag = 1;
                break;
            case 's':
                glob_pktsend_flag = 1;
                break;
            case '?':
            default:
                usage(argv[0]);
                return 0;
        }
    }
    
    ///打开设备
    if (pag_open() < 0)
    {
        printf("### Error : can not open device! ###\n");
        return -1;
    }

    ///创建线程
    glob_loop_flag_thread = 1;
    for (i = 0; i < glob_threads_num; i++)
    {
        sids[i] = i;
        if (pthread_create(&rx_tids[i], NULL, (void *)rx_thread, 
            (void *)&sids[i]) != 0)   
        {
            printf("### Error: Can not creat threads!\n");    
            pag_close();
            return -1;
        }
    }
    
    ///如果不打印报文内容，则打印报文速度
    if (!glob_print_pkt_flag)
    {
        if (pthread_create(&print_tid, NULL, (void *)print_speed_thread,
            NULL) != 0)   
        {
            printf("### Error: Can not creat threads!\n");    
            pag_close();
            return -1;
        }
    }

    ///等待结束信号，或着收到期望的包数 
    sigemptyset(&set);
    sigprocmask(SIG_SETMASK, &set, NULL);
    signal(SIGINT, stop_loop);
    
    gettimeofday(&ts, NULL);   
     
    glob_loop_flag_main = 1;
    while (glob_loop_flag_main) 
    {
        usleep(10);  
        
        all_pkts = 0;
        all_bytes = 0;
        for (i = 0; i < glob_threads_num; i++) 
        {
            all_pkts += glob_pkts[i][0];
        }
        
        if (glob_pkts_num != 0 && glob_pkts_num <= all_pkts)
        {
            break;
        }
    }
            
    gettimeofday(&te, NULL);  
    
    ///停止线程
    glob_loop_flag_thread = 0;
    
    for (i = 0; i < glob_threads_num; i++)
    {
        pthread_join(rx_tids[i], NULL);
    }    
    if (!glob_print_pkt_flag)
    {
        pthread_join(print_tid, NULL);
    }
    
    ///关闭设备
    pag_close();    
    
    ///打印总结信息
    sec = (float)(te.tv_sec - ts.tv_sec) + ((float)(te.tv_usec - ts.tv_usec)) 
        / 1000000.f;   
    printf("\n");        
    printf("Pkts recieved in threads: \n");
    all_pkts = 0;
    all_bytes = 0;
    for(i=0; i<glob_threads_num; i++) 
    {
        printf("\t Thread[%d]:\t <%8lu> pkts, <%8lu M> bytes\n", i,
            glob_pkts[i][0], glob_bytes[i][0] >> 20);
        all_pkts += glob_pkts[i][0];
        all_bytes += glob_bytes[i][0];
    }
    printf("Total : %lu pkts,  %lu bytes\n", all_pkts, all_bytes);
    printf("Speed : < %u > pps, < %u M > bps\n", 
        (int)(all_pkts / sec), (int)((all_bytes >> 17) / sec));
    printf("\n");        

    return 0;
}
