////////////////////////////////////////
///@file tcp_sample.c
///
///@brief pag接口tcp流还原示例代码，根据命令行参数输出不同内容
///
/// 修改记录:
///
///@author liuzhh@dawning.com.cn
///@date 2009-11-29
///
/// 修改内容：建立文档
////////////////////////////////////////


//系统路径下头文件
#include <string.h>  
#include <stdio.h>  
#include <pthread.h>
#include <signal.h>
#include <time.h>
#include <sys/types.h>
#include <netinet/ip.h>

//本地路径下包含文件
#include "libpag.h"

///// 宏定义 ///////

#define THREAD_NUM 4 ///<收包线程数
#define LINK_NUM (1000000 / THREAD_NUM)   ///<每个线程内最大并发的tcp连接数

#define REBIULD_PROTO 0x06

////////// 数据 /////////

int sids[THREAD_NUM];

ulong pkts[THREAD_NUM][16];
ulong bytes[THREAD_NUM][16];
ulong old_pkts[THREAD_NUM];
ulong old_bytes[THREAD_NUM];
ulong all_pkts, all_bytes;
ulong estb_stream_num[THREAD_NUM][16], close_stream_num[THREAD_NUM][16];
ulong total_estb_stream_num, total_close_stream_num;
ulong old_estb_stream_num[THREAD_NUM], old_close_stream_num[THREAD_NUM];

int glob_main_loop_flag = 1;
int glob_thread_loop_flag = 1;

int glob_info_level = 0;

///////   函数列表  ///////////////

///@brief 输出程序帮助信息
///@param cmd 程序命令(输入)
void usage(char* cmd) 
{
    printf("Usage: %s [info_level]", cmd);
    printf(" info_level: level of printing tcp_stream data, more higher more accurate\n");
    printf("   0: default, tcp establishing and closing speed\n");
    printf("   1: tcp_stream content \n");
    printf("   2: user data\n");
    printf("\n");
}

///@brief 打印收包信息的线程，每隔1秒显示一次
///@param ptr 线程建立时的指针，这里没有使用
void print_speed_thread(void *ptr)
{
    int i;
    for(i=0; i<THREAD_NUM; i++)
    {
        old_pkts[i] = 0;
        old_bytes[i] = 0;
        old_estb_stream_num[i] = 0;
        old_close_stream_num[i] = 0;
    }
    
    while(glob_thread_loop_flag)
    {
        sleep(1); 
        printf("\nCurrent speed----- ");
        printf("\n");        
        for(i=0; i<THREAD_NUM; i++) 
        {
            printf("\t Thread[%d]:\t %8lu pps, %8lu M bps  tcp:%lu new %lu close\n", 
                i, 
                pkts[i][0] - old_pkts[i], (bytes[i][0] - old_bytes[i]) >> 17, 
                estb_stream_num[i][0] - old_estb_stream_num[i],
                close_stream_num[i][0] - old_close_stream_num[i]);
            old_pkts[i] = pkts[i][0];
            old_bytes[i] = bytes[i][0];
            old_estb_stream_num[i] = estb_stream_num[i][0];
            old_close_stream_num[i] = close_stream_num[i][0];
        }

    }
}

///@brief 处理流结构，包括tcp和udp流
inline void process_stream(void *pdata)
{
    struct tcp_stream *a_tcp;
    struct half_stream *hs, *server, *client;
    int sid;
    __u8 *p8;
    int i;
    
    a_tcp = (struct tcp_stream *)pdata;
    sid = a_tcp->thread_num;

    ///统计连接新建和关闭计数
    if (a_tcp->nids_state == NIDS_HALF_EST || a_tcp->nids_state == NIDS_JUST_EST)
    {
        estb_stream_num[sid][0]++;
        //关闭新建连接
        //pag_delstream(a_tcp, sid);
    }
    if (a_tcp->nids_state == NIDS_CLOSE || a_tcp->nids_state == NIDS_RESET
        || a_tcp->nids_state == NIDS_TIME_OUT)
    {
        close_stream_num[sid][0]++;
    }
    
    ///统计报文数和字节数
    pkts[sid][0]++;
    if (a_tcp->client.count_new)
    {
        bytes[sid][0] += a_tcp->client.count_new;
    }
    else
    {
        bytes[sid][0] += a_tcp->server.count_new;
    }
    
    server = &(a_tcp->server);
    client = &(a_tcp->client);
    
    ///打印tcp_stream结构体信息
    if (glob_info_level > 0)
    {
        printf("\n-----[sid : %d] [pkt: %d]-----\n", sid, pkts[sid][0]);
        printf("tuple4:%x:%d-->%x:%d, state:%d, dir:%d, app proto:%d\n", 
            a_tcp->addr.saddr, a_tcp->addr.source, a_tcp->addr.daddr, a_tcp->addr.dest,
            a_tcp->nids_state, a_tcp->dir, a_tcp->appproto);
        printf("[server] offset:%d, count:%d, count_new:%d, count_ideal:%d, ",
            server->offset, server->count, server->count_new, server->count_ideal);        
        printf("pktcount:%d, pktcount_new:%d, seq:%u, first_data_seq:%u, ack_seq:%u\n",
            server->pktcount, server->pktcount_new, server->seq, server->first_data_seq, server->ack_seq);
        printf("[client] offset:%d, count:%d, count_new:%d, count_ideal:%d, ",
            client->offset, client->count, client->count_new, client->count_ideal);        
        printf("pktcount:%d, pktcount_new:%d, seq:%u, first_data_seq:%u, ack_seq:%u\n",
            client->pktcount, client->pktcount_new, client->seq, client->first_data_seq, client->ack_seq);
    }
    
    ///打印数据内容，每16字节一行，128字节一段
    if (glob_info_level > 1)
    {
        if (server->pktcount_new)
        {
            hs = server;
        }
        else
        {
            hs = client;
        }
          
        p8 = (__u8 *)hs->data;
        for (i = 0; i < hs->count_new; i++) 
        {
            if(i % 8 == 0)
            {
                printf(" ");
            }
            if(i % 128 == 0)
            {
                printf("/n");
            }
            if (i % 16 == 0)
            {
                printf("\n%03x: ",i / 16);
            }
            printf("%02x ", p8[i]);
        }
        printf("\n");
        
        //缓存部分数据到下个包
        //pag_savedata(a_tcp, 0, 16);
    }
}


///@brief 处理原始ip包
inline void process_ip(int sid, void *pdata)
{
    struct ip *iph;
    
    pkts[sid][0]++;
    
    iph = (struct ip *)pdata;    
    bytes[sid][0] += ntohs(iph->ip_len);
}

///@brief 收流线程
void thread(void *ptr)
{

    int sid;
    struct pktinfo *ppkt;
    
    int i;

    sid = *((int *)ptr);
        
    //process pkts    
    while (glob_thread_loop_flag)
    {
        ppkt = (struct pktinfo *)pag_getstream(sid);
        if (NULL == ppkt)
        {
            usleep(500);
            continue;
        }
        
        //处理stream结构或原始ip包
        if (ppkt->ipproto != 0)
        {    
            process_stream(ppkt->pdata);
        }
        else
        {
            process_ip(sid, ppkt->pdata);
        }  
    }
}

///@brief 响应ctrl+c的信号处理线程
void stop_loop(int sig)
{
    glob_main_loop_flag = 0;
}

int main(int argc, char *argv[])
{
    struct timeval ts, te;
    double sec; 
    ulong all_pkts, all_bytes;
    sigset_t set;
    pthread_t tids[THREAD_NUM],tx_tids[THREAD_NUM];
    pthread_t tid,tx_tid;
    int i;

    //print usage message
    usage(argv[0]);
    if (argc > 2)
    {
        return -1;
    }
    
    if (argc == 2)
    {
        glob_info_level = atoi(argv[1]);
    }
    
    //open a device 
    if (pag_open() < 0)
    {
        printf("### Error : can not open device! ###\n");
        return -1;
    }

    //set tcp stream num in each thread
    //pag_setlinknum(LINK_NUM);
    
    //enable tcp/udp reassembling
    //pag_setrebuild(REBIULD_PROTO);
    
    //create processing threads 
    for (i = 0; i < THREAD_NUM; i++)
    {
        estb_stream_num[i][0] = 0;
        close_stream_num[i][0] = 0;
        
        sids[i] = i;
        if (pthread_create(&tids[i], NULL, (void *)thread, (void *)&sids[i]) != 0)   
        {
            printf("### Error: Can not creat threads!\n");    
            pag_close();
            return -1;
        }
    }  
     
    //create printing thread
    if (glob_info_level == 0)
    {
        if (pthread_create(&tid, NULL, (void *)print_speed_thread, NULL) != 0)   
        {
            printf("### Error: Can not creat threads!\n");    
            pag_close();
            return -1;
        }
    }
    
    //wait for stop signal    
    sigemptyset(&set);
    sigprocmask(SIG_SETMASK, &set, NULL);
    signal(SIGINT, stop_loop);
    
    //loop recieving pkts
    gettimeofday(&ts, NULL);      
    while (glob_main_loop_flag) 
    {
        usleep(500);    
    }  
    gettimeofday(&te, NULL);  
    
    //stop threads
    glob_thread_loop_flag = 0;
    for (i = 0; i < THREAD_NUM; i++)
    {
        pthread_join(tids[i], NULL);
    }
    if (glob_info_level == 0)
    {
        pthread_join(tid, NULL);
    }
    //sleep(2);

    //close device
    pag_close();    

    //print summary
    sec = (float)(te.tv_sec - ts.tv_sec) 
        + ((float)(te.tv_usec - ts.tv_usec)) / 1000000.f;   
    printf("\n");        
    printf("Pkts recieved in threads: \n");
    all_pkts = 0;
    all_bytes = 0;
    for (i = 0; i < THREAD_NUM; i++)
    {
        printf("\t Thread[%d]:\t <%8lu> pkts, <%8lu M> bytes\n", 
        i, pkts[i][0], bytes[i][0] >> 20);
        all_pkts += pkts[i][0];
        all_bytes += bytes[i][0];
    }

    printf("Total : %lu pkts,  %lu bytes\n", all_pkts, all_bytes);
    printf("Speed : < %lu > pps, < %lu M > bps\n", 
        (ulong)(all_pkts / sec), (ulong)((all_bytes >> 17) / sec));
    printf("\n");        
    
    for (i = 0; i < THREAD_NUM; i++)
    {
        printf( "Thread[%d]:  tcp: <%8lu> established,  <%8lu> closed\n", i, 
            estb_stream_num[i][0], close_stream_num[i][0]);
        total_estb_stream_num += estb_stream_num[i][0];
        total_close_stream_num += close_stream_num[i][0];
    }
    printf(" Total established stream : %8lu,  Total closed stream : %8lu\n", 
        total_estb_stream_num, total_close_stream_num );
    printf("Speed : < %lu > new ps, < %lu M > close ps\n", 
        (ulong)(total_estb_stream_num / sec), (ulong)(total_close_stream_num / sec));
    return 0;
}
