////////////////////////////////////////
///@file tx_eth_sample.c
///
///@brief pag接口发ethernet帧示例代码，可以指定发包的长度，源目ip，和发包端口
///
///修改记录:
///
///@author liuzhh@dawning.com.cn
///@date 2010-02-10
///
///建立文档
////////////////////////////////////////

//系统路径下头文件
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>  //getopt
#include <string.h>
#include <pthread.h>
#include <signal.h>
#include <sys/time.h>
#include <linux/types.h>
#include <netinet/ip.h>  //struct ip
#include <netinet/in.h>  //IPPROTO_TCP
#include <arpa/inet.h>  //inet_addr();

#include <libpag.h>

///// 宏定义 ///////

#define MAX_TX_PKT_LEN  3000  ///<可发送的最大报文长度
#define MAX_TX_PORT_NUM 4

////////// 数据 /////////

int glob_loop_flag_main;
int glob_loop_flag_thread;
int glob_pkts_num;
int glob_port_num;
int glob_pkt_len;
int glob_sid;
int glob_port;
int glob_inc_len_flg;
int glob_pkt_low_len;
int glob_pkt_up_len;
int glob_port_rand_flg;
__u32 glob_sip;
__u32 glob_dip;

ulong glob_pkts;
ulong glob_bytes;


///////   函数列表  ///////////////

///@brief 输出程序帮助信息
///@param cmd 程序命令(输入)
void usage(char* cmd) 
{
    printf("Usage: %s ", cmd);
    printf("[-n pkt_num] [-t pkt num] [-p nic_port] [-a add] [-l low_len] [-u up_len] [-s src_ip] [-d dst_ip]\n");
    printf("  -n pkts num to tx, default = continously\n");
    printf("  -t send pkt from set port num randly\n");
    printf("  -p send pkt from set port \n");
    printf("  -a pkt length increase \n");
    printf("  -l pkt lenth, rang[low,up]0 = random, default = 0\n");
    printf("  -u pkt lenth, rang[low,up]0 = random, default = 0\n");printf("  -p nic port used to tx pkt, as well as tx buf id, default = 0\n");
    printf("  -s src ip of pkt, 0 = random, default = 0\n");
    printf("  -d dst ip of pkt, 0 = random, default = 0\n");
    printf("\n");
}

///@brief 打印发包信息的线程，每隔1秒显示一次
///@param ptr 线程建立时的指针，这里没有使用
void print_thread(void *ptr)
{
    ulong old_pkts;
    ulong old_bytes;
    
    old_pkts = 0;
    old_bytes = 0;
    
    while (glob_loop_flag_thread)
    {
        sleep(1); 
        printf("\nCurrent speed :\t %8lu pps, %8lu M bps\n", 
            glob_pkts - old_pkts, 
            (glob_bytes - old_bytes) >> 17);
        old_pkts = glob_pkts;
        old_bytes = glob_bytes;
    }
}

///@brief 构造报文
///@param pkt 指向报文的指针
///@param len 报文长度
///@param proto 报文类型，17为日志包（udp）,6为封堵包（tcp）
inline void build_pkt(void *pkt, int len, int proto, __u32 sip, __u32 dip)
{
    struct iphdr *iph;
    
    ///找到报文ip头
    iph = (struct iphdr *)pkt;
    
    ///修改各字段
    iph->tot_len = htons(len);
    iph->protocol = proto;
    iph->saddr = sip;
    iph->daddr = dip;
}

///@brief 根据全局参数构造报文，并发送出去，收到中断信号或达到发包数后退出
void tx_pkt()
{
    int i, r, len, proto, sid, port;
    __u32 sip, dip;
    void *pkt;
    static int temp;
    __u8 p8[MAX_TX_PKT_LEN] = 
        {
        0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f, 0x01, 0x02, 
        0x03, 0x04, 0x05, 0x06, 0x08, 0x00,  //eth head
        0x45, 0x00, 0x00, 0x92, 0xc7, 0x6b, 0x00, 0x00, 
        0x00, 0x00, 0x00, 0x00, 0x01, 0x01, 0x01, 0x01,
        0x02, 0x02, 0x02, 0x02, //ip head
        0x00, 0x50, 0x07, 0xbb, 0x78, 0x28, 0x73, 0x3e, 
        0x43, 0x23, 0x6a, 0x7c, 0x70, 0x12, 0x16, 0x58, 
        0x76, 0xd4, 0x00, 0x00//tcp head
        };
    
    ///把报文的应用数据设置为递增字节
    for (i = 0 ; i < MAX_TX_PKT_LEN - 54; i++)
    {
        p8[54 + i] = (__u8)i;
    }
    
    srand(time(0));   
    sid = glob_sid;
 
    ///循环，直到线程循环标志清零
    while (glob_loop_flag_main)
    {
        ///--如果达到指定的发送包数，则跳出循环
        if (glob_pkts_num != 0 && glob_pkts_num <= glob_pkts)
        {
            break;
        }
        
        pkt = pag_getsendbuf(sid);
        
        ///--如果没有缓冲区，则循环等待
        if (!pkt) 
        {
            usleep(500);
            continue;
        }
        
        r = rand();
        
        if (glob_pkt_low_len && glob_pkt_up_len)
        {
            if (glob_inc_len_flg)
            {
                len = temp++ % (glob_pkt_up_len - glob_pkt_low_len + 1) + glob_pkt_low_len;
            }
            else
            {
                len = r % (glob_pkt_up_len - glob_pkt_low_len + 1) + glob_pkt_low_len;
            }
        }
        else if (glob_pkt_low_len)
        {
            len = glob_pkt_low_len;
        }
        else
        {
            len = (r % MAX_TX_PKT_LEN) + 54;
        }    
        
        ///--把原始报文拷贝到发送缓冲区，根据参数对报文修改报文内容
        memcpy(pkt - 14, p8, len);
        
        if (glob_sip)
        {
            sip = glob_sip;
        }
        else
        {
            sip = (__u32)r;
        } 
        
        if (glob_dip)
        {
            dip = glob_dip;
        }
        else
        {
            dip = (__u32)(~r);
        } 
        
        proto = IPPROTO_TCP;
        
        ///--构造报文
        build_pkt(pkt, len - 14, proto, sip, dip);

        ///--发送报文
        if (glob_port_rand_flg)
        {
            port = rand() % glob_port_num;
        }
        else
        {
            port = glob_port;
        }
        if (pag_send_eth(port, sid, len) == -1)
        {
            continue;
        }
       
        ///--增加报文计数
        glob_pkts++;
        glob_bytes += len;        
    }
}


///@brief 信号响应函数，收到信号后把全局循环标志无效掉
void stop_loop(int sig)
{
    glob_loop_flag_main = 0;
}

///@brief 主函数
///@return 成功返回0，失败返回-1
int main(int argc, char *argv[])
{
    int c;
    struct timeval ts, te;
    double sec; 
    sigset_t set;
    pthread_t print_tid;
    
    ///打印使用信息
    //usage(argv[0]);
    if (argc > 11)
    {
        return -1;
    }
    
    ///读取参数
    glob_pkts_num = 0;
    glob_pkt_len = 0;
    glob_port = 0;
    glob_sip = 0;
    glob_dip = 0;
    glob_inc_len_flg = 0;
    glob_pkt_low_len = 0;
    glob_pkt_up_len = 0;
    glob_port_rand_flg = 0; 
   
    while ((c = getopt(argc, argv, "n:al:u:p:s:d:t:?")) != -1) 
    {
        switch (c)
        {
            case 'n':
                glob_pkts_num = atoi(optarg);
                break;
            case 'a':
                glob_inc_len_flg = 1;
                break;
            case 'l':
                glob_pkt_low_len = atoi(optarg);
                if (glob_pkt_low_len < 54)
                {
                    glob_pkt_low_len = 54;
                }
                if (glob_pkt_low_len > MAX_TX_PKT_LEN)
                {
                    glob_pkt_low_len = MAX_TX_PKT_LEN;
                }
                break;
            case 'u':
                glob_pkt_up_len = atoi(optarg);
                if (glob_pkt_up_len > MAX_TX_PKT_LEN)
                {
                    glob_pkt_up_len = MAX_TX_PKT_LEN;
                }
                break;
            case 'p':
                glob_port = atoi(optarg);
                break;
            case 's':
                glob_sip = inet_addr(optarg);
                break;
            case 'd':
                glob_dip = inet_addr(optarg);
                break;
            case 't':
                glob_port_rand_flg = 1;
                glob_port_num = atoi(optarg);
                if (glob_port_num > MAX_TX_PORT_NUM)
                {
                    glob_port_num = MAX_TX_PORT_NUM;
                }
                break;
            case '?':
            default:
                usage(argv[0]);
            return 0;
        }
    }
    
    if ((glob_pkt_up_len && glob_pkt_low_len) && (glob_pkt_up_len < glob_pkt_low_len))
    {
        printf("Warning:send len rang[%d,%d]\n", glob_pkt_low_len,glob_pkt_up_len);
        return;
    }
    ///设置发送缓冲区号等于mac端口号
    glob_sid = glob_port;
    
    ///打开设备
    if (pag_open() < 0)
    {
        printf("### Error : can not open device! ###\n");
        return -1;
    }

    ///创建线程，打印报文速度
    glob_loop_flag_thread = 1;
    if (pthread_create(&print_tid, NULL, (void *)print_thread,
            NULL) != 0)   
    {
        printf("### Error: Can not creat threads!\n");    
        pag_close();
        return -1;
    }
    
    ///等待结束信号
    sigemptyset(&set);
    sigprocmask(SIG_SETMASK, &set, NULL);
    signal(SIGINT, stop_loop);
    
    gettimeofday(&ts, NULL);   
    
    ///启动发包
    glob_loop_flag_main = 1;     
    tx_pkt();
    
    gettimeofday(&te, NULL);  
    
    ///停止线程
    glob_loop_flag_thread = 0;
    pthread_join(print_tid, NULL);
      
    ///关闭设备
    pag_close();   
    
    ///打印总结信息
    sec = (float)(te.tv_sec - ts.tv_sec) + ((float)(te.tv_usec - ts.tv_usec)) 
        / 1000000.f;   
    printf("\n");        
    printf("Total sent :\t <%8lu> pkts, <%8lu M> bytes\n", 
        glob_pkts, glob_bytes >> 20);
    printf("Speed : < %u > pps, < %u M > bps\n", 
        (int)(glob_pkts / sec), (int)((glob_bytes >> 17) / sec));
    printf("\n");        

    return 0;
}
